#
#   5B - Level
#

import os
from albow.sound import stop_sound
from compat import set
from part import Part
from wire import Wire, WirePoint
from network import Network
from tyger import Tyger
import settings
from settings import seconds_to_frames, frames_to_seconds
from vector_algebra import point_line_distance

level_end_delay = 5 # seconds
wait_deadline_speed = 240 # times normal speed

RESET = 'reset'
PLAYING = 'playing'
SUCCEEDED = 'succeeded'
FAILED = 'failed'

class Level(object):
	#  game           Game
	#  parts          [Part]
	#  wires          [Wire]
	#  selection      set of Part or Wire
	#  editing_wire   Wire
	#  changed        boolean
	#  pathname       string
	#  state          PLAYING, SUCCEEDED or FAILED
	#  network        Network
	#  focused_part     Part or None
	#  level_end_timer  int or None   frames
	#  start_time_secs  int
	#  deadline_secs    int
	#  current_time     int           frames
	#  deadline         int           frames
	#  tyger         Tyger
	#  is_tutorial   boolean
	#  time_speed    int
	
	level_file_suffix = ".5bl"
	level_file_magic = "5BL"
	level_file_version = 1

	changed = False
	is_tutorial = False
	testing = False
	time_speed = 1
	
	def __init__(self):
		self.parts = []
		self.wires = []
		self.selection = set()
		self.pathname = None
		self.state = None
		self.network = None
		self.focused_part = None
		self.start_time_secs = -1
		self.deadline_secs = -1
		self.current_time = -1
		self.deadline = -1
		self.tyger = Tyger(self)
	
	def reset(self):
		stop_sound()
		self.state = RESET
		self.level_end_timer = None
		self.current_time = self.secs2frms(self.start_time_secs)
		self.deadline = self.secs2frms(self.deadline_secs)
		for part in self.parts:
			part.reset()
		self.tyger.reset()
		if not self.focused_part:
			self.focused_part = self.find_a_green_phone()
		self.time_speed = 1
	
	def secs2frms(self, secs):
		if secs >= 0:
			return seconds_to_frames(secs)
		else:
			return -1
	
	def restart(self):
		self.reset()
		self.build_network()
		if self.tyger.script or self.contains_ending_control_part():
			self.state = PLAYING
		else:
			self.state = SUCCEEDED
	
	def contains_ending_control_part(self):
		for part in self.parts:
			if part.controls_ending:
				return True

	def add_part(self, part):
		self.parts.append(part)
		self.connect_wires()
		self.changed = True
	
	def add_wire(self, wire):
		self.wires.append(wire)
		self.connect_wire(wire)
		self.changed = True
	
	def extend_wire_to_point(self, wp, p):
		#  Returns new endpoint or None.
		wp2 = wp.extend_wire_to_point(p)
		if wp2:
			wire = wp.wire
			if wire.num_points() == 2:
				self.add_wire(wire)
			self.connect_wire(wp.wire)
			self.changed = True
			return wp2
	
	def merge_wires(self, wp1, wp2):
		print "Level: merging wires:" ###
		print "Wire 1:", wp1.wire, wp1.wire.corners ###
		print "Wire 2:", wp2.wire, wp2.wire.corners ###
		assert not wp1.connected_part and not wp2.connected_part
		wire1 = wp1.wire
		wire2 = wp2.wire
		assert wire1 is not wire2
		end1 = wire1.find_end(wp1)
		end2 = wire2.find_end(wp2)
		if end1 == 0:
			wire1.corners.reverse()
		if end2 == -1:
			wire2.corners.reverse()
		wire1.corners.extend(wire2.corners)
		for wp in wire2.corners:
			wp.wire = wire1
		if wire1 not in self.wires:
			self.add_wire(wire1)
		print "Removing:", wire2, wp2.wire.corners ###
		self.wires.remove(wire2)
		print "Result:", wire1, wp1.wire.corners ###
		self.changed = True

	def add_to_selection(self, item):
		self.selection.add(item)
	
	def remove_from_selection(self, item):
		self.selection.remove(item)

	def toggle_selection(self, item):
		if item in self.selection:
			self.remove_from_selection(item)
		else:
			self.add_to_selection(item)

	def clear_selection(self):
		self.selection.clear()
	
	def offset_selection(self, v, fork_points):
		if self.selection:
			for item in self.selection:
				item.offset(v, fork_points)
			self.update_connections_to_selection() #fork_points)
			self.changed = True
	
	def rotflip_selection(self, bits):
		if self.selection:
			for item in self.selection:
				if item.is_rotatable:
					item.orientation ^= bits
			self.update_connections_to_selection() #fork_points = False)
			self.changed = True

	def update_connections_to_selection(self): #, fork_points = False):
		for wire in self.wires_connected_to_items(self.selection):
			wire.extend_connected_endpoints() #fork_points)
		self.connect_wires()
	
	def connect_wires(self):
		for wire in self.wires:
			self.connect_wire(wire)
	
	def wires_connected_to_items(self, items):
		result = set()
		for item in items:
			if isinstance(item, Wire):
				result.add(item)
			elif isinstance(item, WirePoint):
				result.add(item.wire)
			elif isinstance(item, Part):
				for s in item.connected_wire_points:
					for wp in s:
						result.add(wp.wire)
		return result

	def delete_selection(self):
		selection = self.selection.copy()
		for item in selection:
			self.remove(item)
			self.changed = True

	def remove(self, item):
		if isinstance(item, Part):
			if item in self.parts:
				self.parts.remove(item)
			item.disconnect()
			if item is self.focused_part:
				self.focused_part = None
		elif isinstance(item, Wire):
			if item in self.wires:
				self.wires.remove(item)
			item.disconnect()
		elif isinstance(item, WirePoint):
			wire = item.wire
			item.delete()
			if wire.num_points() < 2:
				self.remove(wire)
		if item in self.selection:
			self.selection.remove(item)
	
	def cut_wire(self, wp, delete_segment = False):
		#  Split wire in two at wire point. If delete_segment, removes
		#  the following wire segment. Assumes the point is not an endpoint.
		wire1 = wp.wire
		corners1 = wire1.corners
		i = corners1.index(wp)
		wire2 = Wire(wire1.color_no)
		corners2 = wire2.corners
		if not delete_segment:
			wp2 = WirePoint(wire2, wp.point)
			corners2.append(wp2)
		for wp2 in wire1.corners[i+1:]:
			wp2.wire = wire2
			corners2.append(wp2)
		del corners1[i+1:]
		self.add_wire(wire2)
	
	def delete_wire_segment(self, wp):
		if wp.wire.find_end(wp) is not None:
			wp.delete()
			self.changed = True
		else:
			self.cut_wire(wp, delete_segment = True)

	def find_item(self, p):
		wp = self.find_wire_point(p)
		if wp:
			return wp
		return self.find_part(p)
	
	def find_wire_segment(self, p):
		#  If p is near a wire segment, returns preceding wire
		#  point, else returns None.
		candidates = []
		for wire in self.wires[::-1]:
			wp1 = None
			for wp2 in wire.corners:
				if wp1:
					d = point_line_distance(p, (wp1.point, wp2.point))
					if d is not None and d < 5:
						candidates.append((d, wp1))
				wp1 = wp2
		if candidates:
			return min(candidates)[1]

	def find_wire_point(self, p):
		hits = []
		for wire in self.wires[::-1]:
			for wp in wire.corners:
				#print "Level: find_wire_point: trying", wp.point ###
				v = wp.point - p
				d = max(map(abs, v.coords))
				if d <= 5:
					hits.append((d, wp))
		if hits:
			d, wp = min(hits)
			return wp

	def find_part(self, p):
		for part in self.parts[::-1]:
			if part.contains_global_point(p):
				return part
	
	def find_terminal(self, p):
		part = self.find_part(p)
		if part:
			i = part.find_terminal(p)
			if i is not None:
				return part, i
			else:
				return None

	def connect_wire(self, wire):
		for i in (0, -1):
			wp = wire.corners[i]
			self.connect_wire_point(wp)

	def connect_wire_point(self, wp):
		if not wp.connected_part:
			p = wp.point
			part = self.find_part(p)
			if part:
				terminal_no = part.find_terminal(p)
				if terminal_no is not None:
					wp.connect_to_terminal(part, terminal_no)
	
	def connect_everything(self):
		d = {}
		for part in self.parts:
			for i in xrange(part.num_terminals()):
				p = part.global_terminal_position(i).coords
				d[p] = (part, i)
		for wire in self.wires:
			for j in (0, -1):
				wp = wire.corners[j]
				if not wp.connected_part:
					p = wp.point.coords
					if p in d:
						part, i = d[p]
						wp.connect_to_terminal(part, i)

	def create_from_file(cls, r):
		if r.token <> cls.level_file_magic or not isinstance(r.args, int):
			r.error("Does not seem to be a 5B level file")
		if r.args <> cls.level_file_version:
			r.error("Wrong file version, expected %s, got %s" % (
				cls.level_file_version, r.args))
		r.next_line()
		level = cls()
		level.read_from_file(r)
		return level
	
	create_from_file = classmethod(create_from_file)
	
	def read_from_file(self, r):
		self.pathname = r.path
		if r.looking_at("Time"):
			self.start_time_secs = r.get_arg(int)
			self.deadline_secs = r.get_arg(int)
			r.next_line()
		self.tyger.read_from_file(r)
		r.expect("PARTS")
		while not r.looking_at("END_PARTS"):
			part = Part.create_from_file(r)
			self.parts.append(part)
		r.expect("END_PARTS")
		r.expect("WIRES")
		while not r.looking_at("END_WIRES"):
			wire = Wire.create_from_file(r)
			self.wires.append(wire)
		r.expect("END_WIRES")
		self.connect_everything()

	def write_to_file(self, f):
		f.write("%s %s\n" % (self.level_file_magic, self.level_file_version))
		f.write("Time [%s, %s]\n" % (self.start_time_secs, self.deadline_secs))
		self.tyger.write_to_file(f)
		f.write("PARTS\n")
		for part in self.parts:
			part.write_to_file(f)
		f.write("END_PARTS\n")
		f.write("WIRES\n")
		for wire in self.wires:
			wire.write_to_file(f)
		f.write("END_WIRES\n")

	def build_network(self):
		for part in self.parts:
			part.clear_nodes()
		self.network = Network(self)
		#self.network.dump()

	def begin_frame(self):
		self.network.reset_nodes()
		parts = self.parts
		for part in parts:
			part.source()
		for part in parts:
			part.begin_frame()
			if part.level_ended():
				if self.state == PLAYING:
					print "Level: Ended by part", part, "in state", self.state ###
					self.state = FAILED
					self.tyger.give_up()
					self.level_end_timer = seconds_to_frames(level_end_delay) + 1
		current_time = self.current_time
		deadline = self.deadline
		if deadline > -1 and current_time == deadline:
			self.time_speed = 1
			self.tyger.deadline_reached()
		self.tyger.begin_frame()
		if self.state == PLAYING and self.tyger.finished():
			self.state = SUCCEEDED
			if not self.is_tutorial:
				self.game.record_progress(self.get_filename())
		if current_time > -1:
			current_time += self.time_speed
			if deadline > -1:
				current_time = min(current_time, deadline)
			self.current_time = current_time
		if self.level_end_timer is not None:
			if self.level_end_timer > 0:
				self.level_end_timer -= 1
	
	def playing(self):
		return self.state == PLAYING

	def succeeded(self):
		return self.state == SUCCEEDED

	def failed(self):
		return self.state == FAILED

	def auto_end(self):
		return False
#		timer = self.level_end_timer
#		return timer is not None and timer == 0

	def script_finished(self):
		print "Level: Script finished"
		#if self.state == PLAYING:
		#	self.state = SUCCEEDED

	def get_script(self):
		return self.tyger.get_script()
	
	def set_script(self, x):
		self.tyger.set_script(x)
		self.changed = True

	def find_a_green_phone(self):
		return self.find_a_phone(0)
	
	def find_a_red_phone(self):
		return self.find_a_phone(1)
	
	def find_a_phone(self, color_no):
		from parts import Telephone
		for part in self.parts:
			if isinstance(part, Telephone):
				if part.color_no == color_no:
					return part

	def tyger_active(self):
		return self.tyger.is_active()

	def tyger_thwarted(self):
		return self.tyger.script and self.state == SUCCEEDED

	def tyger_succeeded(self):
		return self.state == FAILED

	def get_name(self):
		path = self.pathname
		if path:
			name = os.path.basename(path)
			return os.path.splitext(name)[0]
		else:
			return "Untitled"
	
	def get_filename(self):
		path = self.pathname
		if path:
			return os.path.basename(path)
		else:
			return None

	def rot_flip(self, part, bits):
		part.orientation ^= bits
		self.changed = True

	def wait_until_deadline(self):
		self.time_speed = wait_deadline_speed
		if self.deadline < 0:
			self.deadline = self.current_time

	def has_tyger_script(self):
		return bool(self.tyger.script)

	def edit_selection_params(self):
		for part in self.selection:
			result = part.edit_params()
			assert isinstance(result, bool)
			if result:
				self.update_connections_to_selection()
				self.changed = True
			return

	def set_start_time_secs(self, x):
		self.start_time_secs = x
		self.changed = True

	def set_deadline_secs(self, x):
		self.deadline_secs = x
		self.changed = True
