#
#   5B - Part
#

from pygame import draw
from albow.resource import get_image
from image import RotatableImage
from vector_algebra import Vector2, Point2, OrientMat2, Transformation2, \
	IdentTran2
from compat import set

class Part(object):
	#  image_name       string
	#  terminal_points  [(int, int)]
	#
	#  position               Vector2          position of centre
	#  orientation            rotcode          rotation/flipping
	#  image                  RotatableImage
	#  connected_wire_points  [set]            indexed by terminal no.
	#  nodes                  [Node]           indexed by terminal no.
	
	connection_image = get_image("connection.png")
	highlight_width = 3
	color_names = []
	image_name = None
	image_border = None
	image = None
	controls_ending = False
	is_rotatable = False
	edit_params_on_create = False

	next_serial = 1
	
	color_no = 0
	orientation = 0
	
	def __init__(self):
		self.connected_wire_points = []
		self.serial = Part.next_serial
		Part.next_serial += 1
		image_name = self.image_name
		if image_name and not self.image:
			self.__class__.image = RotatableImage(get_image(image_name,
				border = self.image_border))
		self.update_terminals()
		self.reset()
		self.clear_nodes()
	
	def set_terminal_points(self, tp):
		self.terminal_points = tp
		self.update_terminals()
	
	def update_terminals(self):
		sets = self.connected_wire_points
		n = self.num_terminals()
		for i in xrange(n, len(sets)):
			for wp in sets[i].copy():
				wp.disconnect()
		del sets[n:]
		while len(sets) < n:
			sets.append(set())
	
	def __repr__(self):
		return "%s#%s" % (self.__class__.__name__, self.serial)
	
	def init_from_catalog(self):
		pass
		
	def reset(self):
		pass
	
	def clear_nodes(self):
		self.nodes = [None] * self.num_terminals()
	
	def find_reachable_terminals(self, terminal_no, result):
		t = (self, terminal_no)
		if t not in result:
			result.add(t)
			for wp in self.connected_wire_points[terminal_no]:
				wp.wire.find_reachable_terminals(result)
	
	def draw(self, surf, t):
		#  Draw with transformation t on surface surf
		image = self.get_rotatable_image()
		image.draw(surf, t * self.position, self.orientation)
	
	def draw_connections(self, surf, t):
		#  Draw solder blobs on connected terminals.
		tt = t * self.local_to_global_transformation()
		connections = self.connected_wire_points
		image = self.connection_image
		r = image.get_rect()
		for i, c in enumerate(self.terminal_points):
			if connections[i]:
				p = tt * Point2(c)
				r.center = p.coords
				surf.blit(image, r)
	
	def get_rotatable_image(self):
		return self.image
	
	def draw_highlight(self, surf, t, color):
		r = self.get_global_rect()
		draw.rect(surf, color, r, self.highlight_width)
	
	def contains_global_point(self, p):
		r = self.get_global_rect()
		return r.collidepoint(p.coords)
	
	def offset(self, v, fork_points):
		self.position += v
		if fork_points:
			for connections in self.connected_wire_points:
				for wp in connections:
					wp.duplicate()
	
	def get_global_rect(self):
		#print "Part: get_global_rect: position =", self.position ###
		return self.get_translated_global_rect(IdentTran2)
	
	def get_translated_global_rect(self, t):
		#  Get rect in global coords translated by t
		r = self.get_oriented_rect()
		c = t * self.position
		r.center = c.coords
		return r
	
	def get_oriented_rect(self):
		image = self.get_rotatable_image()
		return image.get_rect(self.orientation)
	
	def local_to_global_transformation(self):
		return Transformation2(OrientMat2[self.orientation],
			Vector2(self.position.coords))

#	def global_to_local_point(self, gp):
#		return OrientTran2[self.orientation].transpose * gp

	def num_terminals(self):
		return len(self.terminal_points)

	def find_terminal(self, gp):
		for i in xrange(self.num_terminals()):
			tp = self.global_terminal_position(i)
			#print "Part: find_terminal: global terminal point", i, "is", tp ###
			if gp == tp:
				return i

	def global_terminal_position(self, terminal_no):
		v = Vector2(self.terminal_points[terminal_no])
		m = OrientMat2[self.orientation]
		return self.position + m * v

	def disconnect(self):
		for s in self.connected_wire_points:
			for wp in s:
				wp.connected_part = None
				wp.connected_terminal_no = None
			s.clear()
	
	def create_from_file(self, r):
		import parts
		class_name = r.token
		cls = getattr(parts, class_name, None)
		if not cls:
			r.error("Unknown class name '%s'" % class_name)
		part = cls()
		part.read_from_file(r)
		return part
		
	create_from_file = classmethod(create_from_file)

	def read_from_file(self, r):
		self.read_args(r)
		r.next_line()
	
	def read_args(self, r):
		p = r.get_tuple(int, int)
		o = r.get_arg(int)
		self.position = Point2(p)
		self.orientation = o
		self.read_params(r)
	
	def read_params(self, r):
		pass

	def write_to_file(self, f):
		params = [self.position.coords, self.orientation]
		self.get_params(params)
		f.write("%s %s\n" % (self.__class__.__name__, params))
	
	def get_params(self, params):
		pass

	def selectable_at_play_time(self):
		return False

	def key_down(self, e):
		return False
	
	def source(self):
		pass

	def begin_frame(self):
		pass

	def set_node(self, terminal_no, state):
		node = self.nodes[terminal_no]
		if node:
			node.set(state)

	def node_changed(self, terminal_no):
		pass

	def node_state(self, i):
		node = self.nodes[i]
		if node:
			return node.state
		else:
			return 0

	def propagate(self, i, j):
		state1 = self.node_state(i)
		state2 = self.node_state(j)
		if state1 <> state2:
			state = state1 | state2
			self.set_node(i, state)
			self.set_node(j, state)

	def dump_node_states(self):
		print "%s: Nodes = %s" % (self,
			" ".join([str(self.node_state(i)) for i in xrange(self.num_terminals())]))

	def power_across(self, i, j):
		state1 = self.node_state(i)
		state2 = self.node_state(j)
		return state1 ^ state2 == 3

	def level_ended(self):
		return False

	def edit_params(self):
		pass
